/**
 * C++ Class    L3DGECommunicator
 *
 * Revision     1.0a (9th August 2010)
 * Programmer   Michael Allwright
 *
 * Purpose      L3DGECommunicator creates an interface between the LTCPMON and the L3DGEComms Library
 *              by Lucas Parry. The constructor of this class takes the IP address and port of the remote
 *              server, along with the entities file (aka L3DGEhosts.conf) which lists the entities.
 *              it is important that this file matches the file in use on the server for a clean output
 *
 *              Once the object has been constructed, methods for updating and reseting the visual metrics
 *              of the entity are provided
**/

#include <iostream>
#include <sstream>
#include <iomanip>
#include <string>
#include <vector>
#include <map>

#include <cstring>
#include <cstdio>

#include "L3DGECommunicator.h"
#include "l3dgecomms.h"
#include "debug.h"
#include "utilities.h"
#include "tcpmapping.h"
#include "configuration.h"
#include "entityfileparser.h"
#include "entitymapping.h"

/**
 * Method Name  L3DGECommunicator::L3DGECommunicator(const std::string & serverIPAddress, unsigned int serverPort,
 *                                                   const std::string & entityFilename)
 *
 * Parameters   const std::string & serverIPAddress - C++ String containing the L3DGEWorld Server IP Address
 *              unsigned int serverPort             - Server port number
 *              const std::string & entityFilename  - C++ String containing the entity file name.
 *
 * Description  Constructor
**/
L3DGECommunicator::L3DGECommunicator(const std::string & serverIPAddress, unsigned int serverPort, const std::string & entityFilename)
  : entities(entityFilename)
{
    strcpy(addressBuffer, serverIPAddress.c_str());
    valueBuffer[0] = '\0';
    nameBuffer[0] = '\0';
    debug=0;
    setupComms(NULL, 0, addressBuffer, serverPort);
    resetEntities();
}


/**
 * Method Name  void L3DGECommunicator::resetEntities()
 *
 * Description  Iterates accross the entities maps, clearing all TCP Flow assignments and setting the visual metrics to 0
 *              by calling resetEntity(entityUID) in a loop
**/
void L3DGECommunicator::resetEntities()
{
    std::vector<EntityMapping>::const_iterator it;
    for(it = entities.getEntityMappings().begin(); it !=  entities.getEntityMappings().end(); it++)
        resetEntity(it->entityUID);
}


/**
 * Method Name  void L3DGECommunicator::resetEntity
 *
 * Parameters   unsigned int entityUID - UID of the entity to be reset
 *
 * Description  Iterates accross the visual metrics setting the values and rates to 0. also removes the association from
 *              entity mappings list so that the entity can be used
 *
 * TODO / BUG   At this point the entities do not seem to be reseting correctly. also note that removing an entity
 *              association will allow for it to be assigned a new TCP flow. this can be ambiguous during the
 *              visualisation
**/
void L3DGECommunicator::resetEntity(unsigned int entityUID)
{
    for(unsigned int i = 1; i < 6; i++)
    {
        sendMessage(entityUID, i, FIELD_VALUE, "0");
        sendMessage(entityUID, i, FIELD_RATE, "0");
    }
    sendMessage(entityUID, 0, FIELD_NAME, "Inactive Flow Entity");
    entities.removeEntityAssociation(entityUID);
}


/**
 * Method Name  void L3DGECommunicator::updateEntities
 *
 * Parameters   double fromTimeStamp - Update the entities using information from this timestamp forwards.
 *              const std::map<std::string, TcpFlow> & theTCPSessions - Map of the logged TCP flows
 *              const std::vector<TcpMapping> & theVisualMappings) - List of the parameter to metric mappings.
 *
 * Description  Iterates through the various mappings and visual metrics twice. On the first run, data is used to calibrate
 *              the DataScaler object to keep the visual metrics relative to each other. one the second pass through the
 *              the visual metrics are updated.
 *
 * TODO / BUG   Currently the entity reset and TCPFlow removal logic is flawed, as when an entity is removed it can be
 *              replaced with another flow, making the visualisation ambiguous.
**/
void L3DGECommunicator::updateEntities(double fromTimeStamp, const std::map<std::string, TcpFlow> & theTCPSessions,
                       const std::vector<TcpMapping> & theVisualMappings)
{
    /* some more hacky buffers */
    char strtokBuffer[50];
    char msgBuffer[1000];

    std::ostringstream informationWindow;
    std::vector<unsigned int> parameterList;

    std::map<std::string, TcpFlow>::const_iterator flowIter;
    std::vector<TcpMapping>::const_iterator visualIter;

    /** First Pass  - Calibrate the scaler **/
    for(flowIter = theTCPSessions.begin(); flowIter != theTCPSessions.end(); flowIter++)
    {
        if(Configuration::verboseMode == true) utilities::printFlowMapEntity(*flowIter);

        for(visualIter = theVisualMappings.begin(); visualIter != theVisualMappings.end(); visualIter++)
        {
            double calibrationValue;
            parameterList = flowIter->second.getParameterList(visualIter->flowCharacteristic,fromTimeStamp);

            if(Configuration::verboseMode == true)
                std::cerr << "\tFound " << parameterList.size() << " new values for "
                          << Configuration::statisticsModesH[visualIter->statisticsMode] << " of "
                          << Configuration::flowCharacteristicsH[visualIter->flowCharacteristic] << std::endl;

            if(parameterList.size() == 0) continue;
            switch(visualIter->statisticsMode)
            {
                case STATMODE_PARAM_MEDIAN:
                    calibrationValue = utilities::statistics::calcMedian(parameterList);
                    break;
                case STATMODE_PARAM_STDDEV:
                    calibrationValue = utilities::statistics::calcStdDev(parameterList);
                    break;
                case STATMODE_PARAM_MINIMUM:
                    calibrationValue = utilities::statistics::calcMin(parameterList);
                    break;
                case STATMODE_PARAM_MAXIMUM:
                    calibrationValue = utilities::statistics::calcMax(parameterList);
                    break;
                default:
                    continue;
            }
            theScaler.calibrateScaler(calibrationValue, visualIter->flowCharacteristic, visualIter->statisticsMode);
        }
    }

    /** Second Pass  - Scale statistics and update metrics via L3DGEComms **/
    for(flowIter = theTCPSessions.begin(); flowIter != theTCPSessions.end(); flowIter++)
    {
    unsigned int entityUID = entities.setEntityAssociation(flowIter->first);

    // Clear information dialog and write in ID
    informationWindow.str("");

    for(visualIter = theVisualMappings.begin(); visualIter != theVisualMappings.end(); visualIter++)
    {
          double updateValue = 0;
          parameterList = flowIter->second.getParameterList(visualIter->flowCharacteristic,fromTimeStamp);

          /* TODO improve entity reset and flow removal logic logic */
          if(parameterList.size() == 0)
          {
          resetEntity(entityUID);
          continue;
          }

          switch(visualIter->statisticsMode)
          {
          case STATMODE_PARAM_MEDIAN:
              updateValue = utilities::statistics::calcMedian(parameterList);
              break;
          case STATMODE_PARAM_STDDEV:
              updateValue = utilities::statistics::calcStdDev(parameterList);
              break;
          case STATMODE_PARAM_MINIMUM:
              updateValue = utilities::statistics::calcMin(parameterList);
              break;
          case STATMODE_PARAM_MAXIMUM:
              updateValue = utilities::statistics::calcMax(parameterList);
              break;
          default:
              continue;
          }

          informationWindow << '@' << Configuration::flowCharacteristicsH[visualIter->flowCharacteristic]
                << " = " << updateValue;
          strcpy(nameBuffer, Configuration::flowCharacteristicsH[visualIter->flowCharacteristic]);

          updateEntity(entityUID, visualIter->visualisationMetric,
               theScaler.scaleToMetric(updateValue,visualIter->flowCharacteristic, visualIter->statisticsMode,
                            visualIter->visualisationMetric));
    }

    // Create message box string for simulation
    strcpy(strtokBuffer, flowIter->first.c_str());
    strtok(strtokBuffer," ");
    sprintf(msgBuffer, "Local = %s@Remote = %s@", strtok(NULL," "), strtok(NULL," "));

    strcat(msgBuffer, informationWindow.str().c_str());
    sendMessage(entityUID, 0, FIELD_NAME, "Active TCP Connection");
    sendMessage(entityUID, 0, FIELD_VALUE, msgBuffer);
    }
}

/**
 * Method Name  void L3DGECommunicator::updateEntity
 *
 * Parameters   unsigned int entityUID - UID of the entity to update
 *              unsigned int flowCharacteristic - Flow characteristic to be updated (i.e. CWND, RTT)
 *              unsigned int visualisationMetric - Visual Metric to project characteristic upon
 *              double value - Scaled Value to be projected
 *
 * Description  Sends the value 'value' to entity 'entityUID' to update field 'visualisationMetric'
 *
 * TODO / BUG   For this function to work nicely, i.e. updating the text correctly, the class scoped buffers valueBuffer
 *              and nameBuffer need to filled in with correct values before running this method. this is typically done
 *              by calling the updateEntities method instead of this method directly, this function is public to provide
 *              extra flexibility and a means of debugging.
**/
void L3DGECommunicator::updateEntity(unsigned int entityUID, unsigned int visualisationMetric, double value)
{
    std::ostringstream message;

    if(visualisationMetric == VISUAL_METRIC_COLOUR)
    {
    message << std::hex << (unsigned int)value << std::dec;
    }
    else
    {
    message << std::dec << (unsigned int)value;
    }

    // Copy strings hack for a hacky library
    strcpy(valueBuffer, message.str().c_str());

    sendMessage(entityUID, visualisationMetric, FIELD_VALUE, valueBuffer);
    sendMessage(entityUID, visualisationMetric, FIELD_RATE, valueBuffer);
    sendMessage(entityUID, visualisationMetric, FIELD_NAME, nameBuffer);
}
