/**
 * C++ Class    Dynamic Data Scaler
 *
 * Revision     1.0a (9th August 2010)
 * Programmer   Michael Allwright
 *
 * Purpose      The Dynamic Data Scaler is used to scale the statisitics for individual flows
 *              into values which can be projected onto L3DGEWorld entities. Before using the
 *              using the scaler, it must be calibrated with all the data that is to be scaled.
 *              This creates a two pass situation. On the first pass, you must pass all new
 *              statistics data to the calibrateScaler method. On the seconds pass you can run each
 *              value through the scaleToMetric method to get a scaled value.
 *
 * TODO / BUG   Currently the data scaler takes the minimum and maximum values and scales the flow
 *              characteristic's statistic linearly between these values. This has 2 problems, 1.
 *              Dynamic scaling between iterations can give the perspective that network conditions
 *              are changing when metrics are just being rescaled, 2. The maximum and minimum values
 *              are sensitive to noise. - this could be fixed using the 5th & 95th percentiles to
 *              filter away the outliers
**/


#include <iostream>

#include "configuration.h"
#include "datascaler.h"
#include "bounds.h"

/**
 * Method Name  DataScaler::DataScaler()
 *
 * Description  Constructor
**/
DataScaler::DataScaler()
{
    resetScaler();
}

/**
 * Method Name  void DataScaler::resetScaler()
 *
 * Description  Resets the scaler, this generally should not be done during runtime
 *              of the main visualisation loop.
**/
void DataScaler::resetScaler()
{
    parameterBounds.clear();
    std::vector<bounds> initVector;
    for(unsigned int i = 0; i < Configuration::numFlowCharacteristics; i++)
    {
        initVector.push_back(bounds(0,0));
    }
    for(unsigned int i = 0; i < Configuration::numStatisticsModes; i++)
    {
        parameterBounds.push_back(initVector);
    }
}

/**
 * Method Name  void DataScaler::calibrateScaler(double quantity, unsigned int flowChar, unsigned int statMode)
 *
 * Parameters   double quantity - The value or result of a particular statistics function
 *              unsigned int flowChar - The flow characteristic which this number came from
 *              unsigned int statMode - The statisitcs mode which outputted this result
 *
 * Description  Calibrates the scaler to scale characteristic's statistics values in between the given max and min
 *              values.
 *
 * TODO / BUG   Currently very sensitive to noise - change to use 5th and 95th percentiles to filter out this noise
 *              be sure to truncate outliers in the scaleToMetric method once this is fixed.
**/
void DataScaler::calibrateScaler(double quantity, unsigned int flowChar, unsigned int statMode)
{
    if(parameterBounds[statMode][flowChar].maxInitFlag = false)
    {
        parameterBounds[statMode][flowChar].maxInitFlag = true;
        parameterBounds[statMode][flowChar].max = quantity;
    }
    else if(quantity > parameterBounds[statMode][flowChar].max)
        parameterBounds[statMode][flowChar].max = quantity;

    if(parameterBounds[statMode][flowChar].minInitFlag = false)
    {
        parameterBounds[statMode][flowChar].minInitFlag = true;
        parameterBounds[statMode][flowChar].min = quantity;
    }
    else if(quantity < parameterBounds[statMode][flowChar].min)
        parameterBounds[statMode][flowChar].min = quantity;
}

/**
 * Method Name  double DataScaler::scaleToMetric(double quantity, unsigned int flowChar, unsigned int statMode,
 *                                 unsigned int metric)
 *
 * Parameters   double quantity - The value or result of a particular statistics function
 *              unsigned int flowChar - The flow characteristic which this number came from
 *              unsigned int statMode - The statistics mode which outputted this result
 *              unsigned int metric - The target metric which the value is being scaled for
 *
 * Description  Scales 'quantity' to match bounds of a given 'metric' given the 'flowChar' and 'statMode'
 *
**/
double DataScaler::scaleToMetric(double quantity, unsigned int flowChar, unsigned int statMode, unsigned int metric)
{
    double scaled = 0;

    switch(metric)
    {
        case VISUAL_METRIC_SPIN:
            // scale to 0 to +200
            if(parameterBounds[statMode][flowChar].min == parameterBounds[statMode][flowChar].max)
            {
                scaled = 100;
            }
            else
            {
                scaled = (quantity * 200) / (parameterBounds[statMode][flowChar].max - parameterBounds[statMode][flowChar].min);
            }
        break;
        case VISUAL_METRIC_BOUNCE_HEIGHT:
            // scale to 0 to +50
            if(parameterBounds[statMode][flowChar].min == parameterBounds[statMode][flowChar].max)
            {
                scaled = 25;
            }
            else
            {
                scaled = (quantity * 50) / (parameterBounds[statMode][flowChar].max - parameterBounds[statMode][flowChar].min);
            }
        break;
        case VISUAL_METRIC_SCALE:
            // scale to +10 to +30
            if(parameterBounds[statMode][flowChar].min == parameterBounds[statMode][flowChar].max)
            {
                scaled = 20;
            }
            else
            {
                scaled = ((quantity * 20) / (parameterBounds[statMode][flowChar].max - parameterBounds[statMode][flowChar].min)) + 10;
            }
        break;
        case VISUAL_METRIC_ROLL:
            // scale to 0 to 2 (integers only)
            if(parameterBounds[statMode][flowChar].min == parameterBounds[statMode][flowChar].max)
            {
                scaled = 0;
            }
            else
            {
                scaled = ((quantity * 2) / (parameterBounds[statMode][flowChar].max - parameterBounds[statMode][flowChar].min));
                // convert to integers
                if(scaled >= 1.5) scaled = 2;
                else if(scaled >= 0.5) scaled = 1;
                else scaled = 0;
            }
        break;
        case VISUAL_METRIC_COLOUR:
            if(parameterBounds[statMode][flowChar].min == parameterBounds[statMode][flowChar].max)
            {
                scaled = 8388608;
            }
            else
            {
                scaled = (quantity * 16777215) / (parameterBounds[statMode][flowChar].max - parameterBounds[statMode][flowChar].min);
            }
        break;
        case VISUAL_METRIC_BOUNCE_RATE:
            if(parameterBounds[statMode][flowChar].min == parameterBounds[statMode][flowChar].max)
            {
                scaled = 50;
            }
            else
            {
                scaled = (quantity * 100) / (parameterBounds[statMode][flowChar].max - parameterBounds[statMode][flowChar].min);
            }
        break;
        default: scaled = 0;
    }
    if(Configuration::verboseMode == true)
    {
        std::cerr << "\tValue " << quantity << " has been scaled to " << scaled
              << " using bounds (" << parameterBounds[statMode][flowChar].min << ", "
              << parameterBounds[statMode][flowChar].max << ")" << std::endl;
    }
    return scaled;
}
