/**
 * C++ Class    SIFTR Log Parser
 *
 * Revision     1.0a (9th August 2010)
 * Programmer   Michael Allwright
 *
 * Purpose      SIFTR Log Parser is designed to parse log files generated by the SIFTR kernel module (developed by CAIA)
 *              
 **/

#include <iostream>
#include <sstream>

#include "configuration.h"
#include "utilities.h"
#include "siftrlogparser.h"

/**
 * Method Name  SiftrLogParser::SiftrLogParser(const std::string & filename)
 *
 * Parameters   const std::string & filename - C++ string containing the file and path of the SIFTR log
 *
 * Description  Constructor
**/
SiftrLogParser::SiftrLogParser(const std::string & filename) : filename(filename)
{
    int enable_time_secs, enable_time_usecs, scanfArgsRead;
    char siftrVersionRead[10];

    // Initialise class wide enableTime and dataLinesRead variables
    enableTime = 0; dataLinesRead = 0;  moreDataFlag = true;

    // Open the logFile
    logFile.open(filename.c_str());
    loadNextLine();

    scanfArgsRead = sscanf(nextLine.c_str(),
    "enable_time_secs=%d\tenable_time_usecs=%d\tsiftrver=%[^\t]\thz=%d\ttcp_rtt_scale=%d\tsysname=%*s\tsysver=%*d\tipmode=%*d",
    &enable_time_secs, &enable_time_usecs, siftrVersionRead, &kernelTickRate, &tcpRttScale);

    if(scanfArgsRead != 5)
    {
    std::ostringstream errorMsg;
    errorMsg << "Initialisation failed : Parsing first line of "<< filename << " failed (unexpected format encounted).";
    throw badLogFileException(errorMsg.str());
    }
    else
    {
    siftrVersion.assign(siftrVersionRead);

    enableTime = enable_time_secs + enable_time_usecs / (double)1000000;

    // Load the next line into nextLine variable, check, increment counters etc.
    loadNextLine();
    parseNextFlowParameters();
    }
}

/**
 * Method Name  void SiftrLogParser::loadNextLine()
 *
 * Description  Prepares the next line for parsing ahead of time (allows nicer error handling)
**/
void SiftrLogParser::loadNextLine()
{
    if(logFile.is_open() == true)
    {
        if(logFile.fail() == false)
        {
            getline(logFile,nextLine);
            dataLinesRead++;
        }
        else
        {
            std::ostringstream errorMsg;
            errorMsg << "Runtime Error : The SIFTR Log \"" << filename << "\" is not accessible";
            throw fileAccessDeniedException(errorMsg.str());
        }
    }
    else
    {
        std::ostringstream errorMsg;
        errorMsg << "Initialisation failed : The SIFTR Log \"" << filename << "\" could not be found";
        throw fileNotFoundException(errorMsg.str());
    }
}

/**
 * Method Name  void SiftrLogParser::parseNextFlowParameters()
 *
 * Description  Parses the next line / entry of the SIFTR log file, parsing syntax accurate as of version 1.2.2 & 1.2.3
**/
void SiftrLogParser::parseNextFlowParameters()
{
    char localIP[51], remoteIP[51];
    int localPort, remotePort, scanfArgsRead;

    // reinitialise nextFlowParameters
    nextFlowParameters = utilities::initialisedTcpParams;

    // Parse the data!
    scanfArgsRead = sscanf(nextLine.c_str(),"%*c,%*x,%lf,%50[^,],%d,%50[^,],%d,%d,%d,%d,%d,%d,%d,%d,%*d,%d,%d,%*d,%*d,%d,%d,%d,%d,%d,%d",
        &nextFlowParameters.packetTimeStamp, localIP, &localPort, remoteIP, &remotePort,
        &nextFlowParameters.parameter[FLOW_PARAM_SLOWSTARTTHRESHOLD],
        &nextFlowParameters.parameter[FLOW_PARAM_CURRENTCONGESTIONWINDOW],
        &nextFlowParameters.parameter[FLOW_PARAM_BANDWIDTHCONTROLLEDWINDOW],
        &nextFlowParameters.parameter[FLOW_PARAM_CURRENTSENDINGWINDOW],
        &nextFlowParameters.parameter[FLOW_PARAM_CURRENTRECEIVEWINDOW],
        &nextFlowParameters.parameter[FLOW_PARAM_CURRENTSENDINGWINDOWSCALINGFACTOR],
        &nextFlowParameters.parameter[FLOW_PARAM_CURRENTRECEIVINGWINDOWSCALINGFACTOR],
        &nextFlowParameters.parameter[FLOW_PARAM_MAXSEGMENTSIZE],
        &nextFlowParameters.parameter[FLOW_PARAM_RTTESTIMATE],
        &nextFlowParameters.parameter[FLOW_PARAM_RETRANSMISSIONTIMEOUTLENGTH],
        &nextFlowParameters.parameter[FLOW_PARAM_SOCKETSENDBUFFERSIZE],
        &nextFlowParameters.parameter[FLOW_PARAM_BYTESINSOCKETSENDBUFFER],
        &nextFlowParameters.parameter[FLOW_PARAM_SOCKETRECEIVEBUFFERSIZE],
        &nextFlowParameters.parameter[FLOW_PARAM_BYTESINSOCKETRECEIVEBUFFER],
        &nextFlowParameters.parameter[FLOW_PARAM_INFLIGHTBYTES]);

    // Check if the line parsed correctly
    if(scanfArgsRead != 20)
    {
        moreDataFlag = false;
        nextFlowParameters = utilities::initialisedTcpParams;
    }
    else
    {
        nextFlowParameters.tcpFlowID.assign(utilities::getTcpFlowID(localIP, localPort, remoteIP, remotePort));
        loadNextLine(); // Load the next line into nextLine variable, check, increment counters etc.
    }
}

/**
 * Method Name  TcpParams SiftrLogParser::getNextFlowParameters()
 *
 * Description  Returns the parsed data
 *
 * Returns      A TcpParams structure (see tcpparams.h), a structure containing all the parsed parameters.
**/
TcpParams SiftrLogParser::getNextFlowParameters()
{
    if(moreDataFlag == false)
    {
        std::ostringstream errorMsg;
        errorMsg << "Parsing line " << dataLinesRead << " of " << filename << " failed (unexpected format encounted).";
        throw noMoreDataException(errorMsg.str());
    }
    else
    {
        TcpParams returnFlow =  nextFlowParameters;
        parseNextFlowParameters();
        return returnFlow;
    }
}

